//-----------------------------------------------------------------------------
//
//                      XMM Science Analysis System (SAS)
//                     (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//  CalOalUtility.cc --- utility functions for CAL/OAL
//
//  DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-11-04: 1.0.0: - some function which used to be inlined
//		UL, 1999-11-10: 1.1.0: - embedded everything in namespace
//		UL, 2001-04-09: 1.2.0: - fillVector() from vector column now also
//								 works for variable length vectors
//		UL, 2001-12-05: 1.3.0: - added template specialization
//								 optionalAttribute<string>
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/caloalutils/src/CalOalUtility.cc,v 1.13 2002/05/24 12:46:05 ulammers Exp $
//
#ifndef CAL_OAL_UTILITY
#define CAL_OAL_UTILITY

#include "stringutil.h"
#include "CalOalUtility.h"

namespace CalOalUtility {

std::string stringColumnRow(const ColumnData *cd, unsigned long row)
{
	unsigned long len = cd->size()[0];
	if ((cd->stringData() + cd->position(row) + len - 1)[0] != '\0') {
		char s[len+1];
		strncpy(s, cd->stringData() + cd->position(row), len);
		s[len] = '\0';
		return std::string(s);
    }

	return std::string(cd->stringData() + cd->position(row));
}

void stringColumnRowSet(const ColumnData *cd, unsigned long row,
					    const std::string &str)
{
	unsigned long len = cd->size()[0];
	strncpy(cd->stringData() + cd->position(row), str.c_str(), len);
}

template<class T>
T optionalAttribute(const Block *block, const std::string &attName,
					T defaultValue)
{
	if (block->hasAttribute(attName))
		return toNumber<T>(block->attribute(attName)->asString());

	return defaultValue;
} // optionalAttribute()

//	instantiations
#define INST(type) \
template type optionalAttribute(const Block *, const std::string &, type);

INST(float)
INST(double)
INST(int)
INST(short)
INST(unsigned)

#undef INST

template<> std::string optionalAttribute<std::string>(const Block *block,
											const std::string &attName,
											const std::string defaultValue)
{
	if (block->hasAttribute(attName))
		return block->attribute(attName)->asString();

	return defaultValue;
} // optionalAttribute()


template<class T>
std::vector<T> &fillVector(const Table *table, const std::string &colName,
						   std::vector<T> &v)
{
	ColumnData *cd = table->column(colName)->data();
	unsigned N = table->rows();
	v.resize(N);

	#define CODE(type, funcname) {			\
		type *dv = cd->funcname();			\
		for (unsigned n=0; n<N; ++n)		\
			v[n] = static_cast<T>(dv[n]);	\
	}

	switch (cd->parent()->dataType()) {
		case Dal::Bool	: CODE(bool8, boolData); break;
		case Dal::Int8	: CODE(int8, int8Data); break;
		case Dal::Uint16: CODE(uint16, uint16Data); break;
		case Dal::Int16	: CODE(int16, int16Data); break;
		case Dal::Uint32: CODE(uint32, uint32Data); break;
		case Dal::Int32	: CODE(int32, int32Data); break;
		case Dal::Real32: CODE(real32, real32Data); break;
		case Dal::Real64: CODE(real64, real64Data); break;
		default:
			delete cd;
			errstr<<"Unexpected type in fillVector()"
				  <<Sas::error("UnExpectedType");
			break;
	}
	delete cd;

	#undef CODE

	return v;
} // fillVector()


template<class T>
std::vector<std::vector<T> > &fillMatrix(const Table *table,
										 const std::string &colName,
										 std::vector<std::vector<T> > &v)
{
	ColumnData *cd = table->column(colName)->data();
	unsigned N = table->rows(),
			 Ne = cd->elements();
	v.resize(Ne);
	for (unsigned n=0; n<Ne; ++n)
		v[n].resize(N);

	#define CODE(type, funcname) {								\
		type *dv = cd->funcname();								\
		for (unsigned n=0; n<N; ++n)							\
			v[i][n] = static_cast<T>(dv[cd->position(n, i)]);	\
	}

	for (unsigned i=0; i<Ne; ++i)
		switch (cd->parent()->dataType()) {
			case Dal::Bool	: CODE(bool8, boolData); break;
			case Dal::Int8	: CODE(int8, int8Data); break;
			case Dal::Uint16: CODE(uint16, uint16Data); break;
			case Dal::Int16	: CODE(int16, int16Data); break;
			case Dal::Uint32: CODE(uint32, uint32Data); break;
			case Dal::Int32	: CODE(int32, int32Data); break;
			case Dal::Real32: CODE(real32, real32Data); break;
			case Dal::Real64: CODE(real64, real64Data); break;
			default:
				delete cd;
				errstr<<"Unexpected type in fillVector()"
					  <<Sas::error("UnExpectedType");
				break;
		}

	delete cd;

	#undef CODE

	return v;
} // fillMatrix()

//
//	fill vector from contents of fixed/variable length vector column
//
//	access to fixed/variable-length vectors columns should be homogenized
//	in the Dal - this should make the below look more elegant
//
template<class T>
std::vector<T> &fillVector(const Table *table, const std::string &colName,
						   unsigned row, std::vector<T> &v)
{
	const Column *col = table->column(colName);
	bool isFixed = col->cellType() == Column::Fixed;

	ColumnData *cdFixedCol = isFixed ? col->data() : 0;
	CellData *cdVarCol = !isFixed ? col->cellData(row) : 0;
	const Data *cd = isFixed ? static_cast<const Data *>(cdFixedCol)
							 : static_cast<const Data *>(cdVarCol);
	unsigned N = isFixed ? cdFixedCol->elements() : cdVarCol->size();
	v.resize(N);

	#define CODE(type, funcname) {										\
		type *dv = cd->funcname();										\
		for (unsigned n=0; n<N; ++n) {									\
			unsigned pos = isFixed ? cdFixedCol->position(row, n) : n;	\
			v[n] = static_cast<T>(dv[pos]);									\
		}																\
	}

	switch (col->dataType()) {
		case Dal::Bool	: CODE(bool8, boolData); break;
		case Dal::Int8	: CODE(int8, int8Data); break;
		case Dal::Uint16: CODE(uint16, uint16Data); break;
		case Dal::Int16	: CODE(int16, int16Data); break;
		case Dal::Uint32: CODE(uint32, uint32Data); break;
		case Dal::Int32	: CODE(int32, int32Data); break;
		case Dal::Real32: CODE(real32, real32Data); break;
		case Dal::Real64: CODE(real64, real64Data); break;
		default:
			delete cd;
			errstr<<"Unexpected type in fillVector()"
				  <<Sas::error("UnExpectedType");
			break;
	}
	delete cd;
	#undef CODE

	return v;
} // fillVector()

#define INST(type) \
template std::vector<type> &fillVector(const Table *, const std::string &, \
				std::vector<type>&);

INST(float)
INST(double)
INST(int)
INST(short)
INST(unsigned)
INST(unsigned short)

#undef INST

#define INST(type) \
template std::vector<std::vector<type> >&fillMatrix(const Table *, \
			const std::string &, std::vector<std::vector<type> > &);

INST(float)
INST(double)
INST(int)
INST(short)
INST(unsigned)
INST(unsigned short)

#undef INST

#define INST(type) \
template std::vector<type> &fillVector(const Table *, const std::string &,	\
								  unsigned, std::vector<type>&);

INST(float)
INST(double)
INST(int)
INST(short)
INST(unsigned)
INST(unsigned short)

#undef INST

std::vector<std::string> &fillVector(const Table *table,
									 const std::string &colName,
									 std::vector<std::string> &v)
{
	ColumnData *cd = table->column(colName)->data();
	unsigned N = table->rows();
	v.resize(N);
	for (unsigned n=0; n<N; ++n)
		v[n] = stringColumnRow(cd, n);

	return v;
} // fillVector()

}

#endif
