//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Celestial.cc --- matrix/vector operations in the celestial frame
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1997-11-13: 1.0.0: - skeleton
//		UL, 1998-08-18: 1.1.0: - normalize angles in AttitudeMatrix constructor
//		UL, 1999-07-01: 1.2.0: - added construction of AttitudeMatrix from
//								 body frame unit vectors
//							   - added unary +/- operator for CartesianVectors
//		UL, 1999-08-04: 1.3.0: - removed inline declaration from some methods
//								 as they were causing linkage problems on
//								 Digital Unix
//							   - added couple of methods related to 1-2-3
//								 Euler angles (credit to JBa)
//		UL, 1999-11-08: 1.4.0: - made EulerAngles a class
//		UL, 2000-03-21: 1.3.0: - added EquatorialDirection constructor
//		UL, 2000-03-31: 1.4.0: - normalize R.A. in construction of
//								 EquatorialDirection from unit vector
//		UL, 2001-03-02: 1.5.0: - added RightAscension/Declination classes
//		UL, 2001-05-11: 1.6.0: - added implementation of WCS related classes
//		UL, 2001-06-04: 1.6.0: - added implementation of
//									EquatorialDirection::toEpoch
//									EquatorialDirection::toApparentDirection
//		UL, 2003-07-04: 1.7.0: - RightAscension/Declination::toString():
//								 in gcc-3.x d/f output format for reals changed
//								 from fixed to scientific; work-around
//								 implemented avoiding real numbers
//		UL, 2003-07-07: 1.8.0: - fixed minor problem introduced with 1.7.0
//							   - TangentialGnomonicProjection::toNativeSystem:
//								 added code to guard against atan2(0,0)
//
//	AUTHOR:
//		U. Lammers, Thu Nov 13 10:42:10 WET 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: Representations of Celestial Coordinates in FITS, Greisen,
//			 Calabreta, 1994
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/caloalutils/src/Celestial.cc,v 1.24 2003/11/20 11:37:16 jhoar Exp $
//
#include <sstream>
#include <iomanip>
#include <cfloat>

#include "Celestial.h"
#include "common.h"
#include "stringutil.h"
#include "errstr.h"

using namespace std;

AttitudeMatrix EulerAngles::toMatrix() const
{
	return AttitudeMatrix(*this);
} // EulerAngles::toMatrix()


std::ostream &operator<<(std::ostream &os, const EulerAngles &eulerAngles)
{
	os<<"Euler-";
	switch (eulerAngles.type()) {
		case EulerAngles::Euler123:	os<<"123"; break;
		case EulerAngles::Euler313:	os<<"313"; break;
		case EulerAngles::Euler321:	os<<"321"; break;
		default: break;
	}
	os<<"(phi="<<setprecision(9)<<eulerAngles.phi() * RAD_TO_DEG
	  <<", theta="<<setprecision(9)<<eulerAngles.theta() * RAD_TO_DEG
	  <<", psi="<<setprecision(9)<<eulerAngles.psi() * RAD_TO_DEG<<')';

	return os;
} // operator<<


CartesianVector::CartesianVector(double x, double y, double z)
{
	_v[0] = x;
	_v[1] = y;
	_v[2] = z;
} // CartesianVector::CartesianVector


CartesianVector &CartesianVector::operator*=(double scalar)
{
	for (unsigned i=0; i<3; _v[i++] *= scalar);

	return *this;
} // CartesianVector::operator*=


CartesianVector &CartesianVector::operator/=(double scalar)
{
	assert(scalar != 0.);

	for (unsigned i=0; i<3; _v[i++] /= scalar);

	return *this;
} // CartesianVector::operator/=


CartesianVector &CartesianVector::operator+=(const CartesianVector &v)
{
	for (unsigned i=0; i<3; ++i)
		_v[i] += v._v[i];

	return *this;
} // CartesianVector::operator+=


CartesianVector &CartesianVector::operator-=(const CartesianVector &v)
{
	for (unsigned i=0; i<3; ++i)
		_v[i] -= v._v[i];

	return *this;
}

CartesianVector operator+(const CartesianVector &v)
{
	return v;
}

CartesianVector operator-(const CartesianVector &v)
{
	return CartesianVector(-v.x(), -v.y(), -v.z());
}


CartesianVector operator+(const CartesianVector &v1, const CartesianVector &v2)
{
	CartesianVector v = v1;

	return v += v2;
} // CartesianVector::operator+


CartesianVector operator-(const CartesianVector &v1, const CartesianVector &v2)
{
	CartesianVector v = v1;

	return v -= v2;
} // CartesianVector::operator+


CartesianVector operator*(const CartesianVector &v1, double scalar)
{
	CartesianVector v = v1;

	return v *= scalar;
} // CartesianVector::operator*

CartesianVector operator*(double scalar, const CartesianVector &v)
{
	return operator*(v, scalar);
}

CartesianVector operator/(const CartesianVector &v1, double scalar)
{
	CartesianVector v = v1;

	return v /= scalar;
} // CartesianVector::operator*


//	scalar product
double operator*(const CartesianVector &v1, const CartesianVector &v2)
{
	return v1._v[0]*v2._v[0] + v1._v[1]*v2._v[1] +
		   v1._v[2]*v2._v[2];
} // operator*


CartesianVector crossProd(const CartesianVector &v1, const CartesianVector &v2)
{
	return CartesianVector(v1._v[1]*v2._v[2]-v1._v[2]*v2._v[1],
						   v1._v[2]*v2._v[0]-v1._v[0]*v2._v[2],
						   v1._v[0]*v2._v[1]-v1._v[1]*v2._v[0]);
} // crossProduct()



//	in/out
std::ostream &operator<<(std::ostream &os, const CartesianVector &v)
{
	os<<"("<<v.x()<<"/"<<v.y()<<"/"<<v.z()<<")";

	return os;
} // operator<<


Plane::Plane(const CartesianVector &p1, const CartesianVector &p2,
			 const CartesianVector &p3) : _origin(p1)
{
	double nx = p1.y() * (p2.z()-p3.z()) +
				p2.y() * (p3.z()-p1.z()) + p3.y() * (p1.z()-p2.z()),
		   ny = p1.z() * (p2.x()-p3.x()) +
				p2.z() * (p3.x()-p1.x()) + p3.z() * (p1.x()-p2.x()),
		   nz = p1.x() * (p2.y()-p3.y()) +
				p2.x() * (p3.y()-p1.y()) + p3.x() * (p1.y() - p2.y());

	if (nx == 0. && ny == 0. && nz == 0.)
		errstr<<"Cannot define plane with three colinear points ("
			  <<p1<<", "<<p2<<", "<<p3
			  <<Sas::fatal("PointsColinear");

	_norm = CartesianVector(nx, ny, nz).normalize();
} // Plane()


AttitudeMatrix::AttitudeMatrix
						(double a00, double a01, double a02,
						 double a10, double a11, double a12,
						 double a20, double a21, double a22)
{
	_matrix[0][0] = a00, _matrix[0][1] = a01; _matrix[0][2] = a02;
	_matrix[1][0] = a10, _matrix[1][1] = a11; _matrix[1][2] = a12;
	_matrix[2][0] = a20, _matrix[2][1] = a21; _matrix[2][2] = a22;

	_eulerValid = false;
} // AttitudeMatrix::AttitudeMatrix


AttitudeMatrix::AttitudeMatrix(const CartesianVector &plusX,
							   const CartesianVector &plusY,
							   const CartesianVector &plusZ)
{
	_matrix[0][0] = plusX.x();
	_matrix[0][1] = plusX.y();
	_matrix[0][2] = plusX.z();

	_matrix[1][0] = plusY.x();
	_matrix[1][1] = plusY.y();
	_matrix[1][2] = plusY.z();

	_matrix[2][0] = plusZ.x();
	_matrix[2][1] = plusZ.y();
	_matrix[2][2] = plusZ.z();

	_eulerValid = false;
} // AttitudeMatrix::AttitudeMatrix


// Spacecraft Attitude Detrmination and Control, Vol 73, pag 413, eq (12-7a)
AttitudeMatrix::AttitudeMatrix(const CartesianVector & axis, double a)
{
	CartesianVector e(axis);
	e.normalize();

	double cosa         = std::cos(a);
	double sina         = std::sin(a);
	double one_min_cosa = 1 - cosa;
	double e1           = e.x();
	double e2           = e.y();
	double e3           = e.z();

	_matrix[0][0] = cosa + e1*e1*one_min_cosa;
	_matrix[0][1] = e1*e2*one_min_cosa + e3*sina;
	_matrix[0][2] = e1*e3*one_min_cosa - e2*sina;

	_matrix[1][0] = e1*e2*one_min_cosa - e3*sina;
	_matrix[1][1] = cosa + e2*e2*one_min_cosa;
	_matrix[1][2] = e2*e3*one_min_cosa + e1*sina;

	_matrix[2][0] = e1*e3*one_min_cosa + e2*sina;
	_matrix[2][1] = e2*e3*one_min_cosa - e1*sina;
	_matrix[2][2] = cosa + e3*e3*one_min_cosa;

	_eulerValid = false;  
} // AttitudeMatrix::AttitudeMatrix()



AttitudeMatrix::AttitudeMatrix(double phi, double theta, double psi,
							   EulerAngles::Type eulerType)
{
	init(phi, theta, psi, eulerType);
} // AttitudeMatrix::AttitudeMatrix()


AttitudeMatrix::AttitudeMatrix(const EulerAngles &ea)
{
	init(ea.phi(), ea.theta(), ea.psi(), ea.type());
} // AttitudeMatrix::AttitudeMatrix()


void AttitudeMatrix::init(double phi, double theta, double psi,
						  EulerAngles::Type eulerType)
{
	_euler = EulerAngles(std::fmod(phi, M_TWO_PI), std::fmod(theta, M_TWO_PI),
						 std::fmod(psi, M_TWO_PI), eulerType);
	_eulerValid = true;

	double	sinPhi		= std::sin(phi),
			cosPhi		= std::cos(phi),
			sinTheta	= std::sin(theta),
			cosTheta	= std::cos(theta),
			sinPsi		= std::sin(psi),
			cosPsi		= std::cos(psi);

	switch (eulerType) {
		case EulerAngles::Euler123:
			//	Spacecraft Attitude Determination and Control, Vol 73,
			//	pag 764, Table 1
			_matrix[0][0] = cosPsi*cosTheta;
			_matrix[0][1] = cosPsi*sinTheta*sinPhi + sinPsi*cosPhi;
			_matrix[0][2] = -cosPsi*sinTheta*cosPhi + sinPsi*sinPhi;
			_matrix[1][0] = -sinPsi*cosTheta;
			_matrix[1][1] = -sinPsi*sinTheta*sinPhi + cosPsi*cosPhi;
			_matrix[1][2] = sinPsi*sinTheta*cosPhi + cosPsi*sinPhi;
			_matrix[2][0] = sinTheta;
			_matrix[2][1] = -cosTheta*sinPhi;
			_matrix[2][2] = cosTheta*cosPhi;
			break;
		case EulerAngles::Euler313:
			_matrix[0][0] = cosPsi*cosPhi - cosTheta*sinPsi*sinPhi;
			_matrix[0][1] = cosPsi*sinPhi + cosTheta*sinPsi*cosPhi;
			_matrix[0][2] = sinTheta*sinPsi;
			_matrix[1][0] = -sinPsi*cosPhi - cosTheta*cosPsi*sinPhi;
			_matrix[1][1] = -sinPsi*sinPhi + cosTheta*cosPsi*cosPhi;
			_matrix[1][2] = sinTheta*cosPsi;
			_matrix[2][0] = sinTheta*sinPhi;
			_matrix[2][1] = -sinTheta*cosPhi;
			_matrix[2][2] = cosTheta;
			break;
		case EulerAngles::Euler321:
			_matrix[0][0] = cosTheta*cosPhi;
			_matrix[0][1] = cosTheta*sinPhi;
			_matrix[0][2] = -sinTheta;
			_matrix[1][0] = -cosPsi*sinPhi + sinPsi*sinTheta*cosPhi;
			_matrix[1][1] = cosPsi*cosPhi + sinPsi*sinTheta*sinPhi;
			_matrix[1][2] = sinPsi*cosTheta;
			_matrix[2][0] = sinPsi*sinPhi + cosPsi*sinTheta*cosPhi;
			_matrix[2][1] = -sinPsi*cosPhi + cosPsi*sinTheta*sinPhi;
			_matrix[2][2] = cosPsi*cosTheta;
		break;
	}
} // AttitudeMatrix::AttitudeMatrix


AttitudeMatrix AttitudeMatrix::inverse() const
{
	AttitudeMatrix m = *this;

	m._matrix[0][1] = _matrix[1][0];
	m._matrix[0][2] = _matrix[2][0];
	m._matrix[1][2] = _matrix[2][1];
	m._matrix[1][0] = _matrix[0][1];
	m._matrix[2][0] = _matrix[0][2];
	m._matrix[2][1] = _matrix[1][2];

	return m;
} // AttitudeMatrix::inverse


EulerAngles AttitudeMatrix::toEulerAngles(EulerAngles::Type eulerType)
{
	if (eulerType == EulerAngles::Euler123)
		return toEuler123();
	if (eulerType == EulerAngles::Euler321)
		return toEuler321();

	return toEuler313();
} // AttitudeMatrix::toEulerAngles()


EulerAngles AttitudeMatrix::toEuler313()
{
	EulerAngles e;

	if (_eulerValid) {
		//	Euler-angle representation is valid
		e = _euler;
		return e;
	}

	//	calculate Euler angles from direction cosine matrix
	if (_matrix[2][2] > 1.-EPSILON) {
		//
		//	only rotation around z
		//
		_euler.theta(0.);
		_euler.phi(std::atan2(_matrix[0][1], _matrix[0][0]));
		_euler.psi(0.);
	} else if (_matrix[2][2] < -1.+EPSILON) {
		//      
		//	flip w/ y and rot around z
		//
        _euler.theta(M_PI);
		_euler.phi(std::atan2(-_matrix[0][1], -_matrix[0][0]));
		_euler.psi(0.);
	} else {
		_euler.theta(std::acos(_matrix[2][2]));

		if (std::fabs(_matrix[2][0]) > EPSILON ||
			std::fabs(_matrix[2][1]) > EPSILON)
			_euler.phi(std::atan2(_matrix[2][0], -_matrix[2][1]));
		else {
			_euler.phi(0.);

			//	illegal matrix
			return e;
		}

		if (std::fabs(_matrix[0][2]) > EPSILON ||
			std::fabs(_matrix[1][2]) > EPSILON)
			_euler.psi(std::atan2(_matrix[0][2], _matrix[1][2]));
		else { 
			_euler.psi(0.);

			//	illegal matrix
			return e;
		}
	}

	normalize(e);
	_eulerValid = true;
	e = _euler;

	return e;
} // AttitudeMatrix::toEuler313()


EulerAngles AttitudeMatrix::toEuler321()
{
	EulerAngles e;

	e.theta(-std::asin(_matrix[0][2]));
	e.phi(std::atan2(_matrix[0][1], _matrix[0][0]));
	e.psi(std::atan2(_matrix[1][2], _matrix[2][2]));
	e.type(EulerAngles::Euler321);
	normalize(e);

	return e;
} // AttitudeMatrix::toEuler321()


EulerAngles AttitudeMatrix::toEuler123()
{
	EulerAngles e;

	e.theta(std::asin(_matrix[2][0]));
	e.phi(std::atan2(-_matrix[2][1],_matrix[2][2]));
	e.psi(std::atan2(-_matrix[1][0],_matrix[0][0]));
	e.type(EulerAngles::Euler123);
	normalize(e);

	return e;
} // AttitudeMatrix::toEuler123()



void AttitudeMatrix::normalize(EulerAngles &eulerAngles) const
{
	if (eulerAngles.phi() < 0.)
		eulerAngles.phi(eulerAngles.phi() + M_TWO_PI);
	if (eulerAngles.psi() < 0.)
		eulerAngles.psi(eulerAngles.psi() + M_TWO_PI);
} // AttitudeMatrix::normalize()




std::ostream &operator<<(std::ostream &os, const AttitudeMatrix &attMat)
{
	os.precision(4);
	os<<(*attMat.data())[0][0]<<" "<<(*attMat.data())[0][1]<<" "
		<<(*attMat.data())[0][2]<<std::endl
	  <<(*attMat.data())[1][0]<<" "<<(*attMat.data())[1][1]<<" "
		<<(*attMat.data())[1][2]<<std::endl
	  <<(*attMat.data())[2][0]<<" "<<(*attMat.data())[2][1]<<" "
		<<(*attMat.data())[2][2]<<std::endl;

	return os;
}

CartesianVector operator*(const AttitudeMatrix &m, const CartesianVector &v)
{
	CartesianVector res(m._matrix[0][0]*v.x() +
						m._matrix[0][1]*v.y() + m._matrix[0][2]*v.z(),
						m._matrix[1][0]*v.x() +
						m._matrix[1][1]*v.y() + m._matrix[1][2]*v.z(),
						m._matrix[2][0]*v.x() +
						m._matrix[2][1]*v.y() + m._matrix[2][2]*v.z());

	return res;
} // operator*


AttitudeMatrix operator*(const AttitudeMatrix& a, const AttitudeMatrix &b)
{
	#define A	a._matrix
	#define B	b._matrix

	return AttitudeMatrix(
		A[0][0]*B[0][0] + A[0][1]*B[1][0] + A[0][2]*B[2][0],
		A[0][0]*B[0][1] + A[0][1]*B[1][1] + A[0][2]*B[2][1],
		A[0][0]*B[0][2] + A[0][1]*B[1][2] + A[0][2]*B[2][2],
		A[1][0]*B[0][0] + A[1][1]*B[1][0] + A[1][2]*B[2][0],
		A[1][0]*B[0][1] + A[1][1]*B[1][1] + A[1][2]*B[2][1],
		A[1][0]*B[0][2] + A[1][1]*B[1][2] + A[1][2]*B[2][2],
		A[2][0]*B[0][0] + A[2][1]*B[1][0] + A[2][2]*B[2][0],
		A[2][0]*B[0][1] + A[2][1]*B[1][1] + A[2][2]*B[2][1],
		A[2][0]*B[0][2] + A[2][1]*B[1][2] + A[2][2]*B[2][2]);

	#undef A
	#undef B
} // AttitudeMatrix * AttitudeMatrix


//
//	---( SpacecraftAttitude )------------------------------------------------
//
AttitudeMatrix SpacecraftAttitude::toAttitudeMatrix() const
{
	return AttitudeMatrix(_ra + M_PI,
							//	because _ra is Right Acension of viewing
							//	direction = -X axis of S/C frame:
							//	_ra + Pi is first Euler angle in 3-2-1
							//	sequence
						  _dec,
							//	declination of of viewing direction is
							//	directly second Euler angle in 3-2-1
							//	sequence
						  _apos,
							//	the astronomical position angle is the third
							//	Euler angle in the 3-2-1 sequence; see [3]
							//	for details
						  EulerAngles::Euler321);
} // SpacecraftAttitude::toAttitudeMatrix()


EulerAngles SpacecraftAttitude::toEulerAngles() const
{
	EulerAngles e(_ra + M_PI, _dec, _apos, EulerAngles::Euler321);

	return EulerAngles(e);
} // SpacecraftAttitude::toEulerAngles() const


std::ostream &operator<<(std::ostream &os, const SpacecraftAttitude &att)
{
	os<<"SpacecraftAttitude("
	  <<RightAscension(att.ra())<<"/"
	  <<Declination(att.dec())<<"/"
	  <<RAD_TO_DEG*att.apos()<<"deg)";

	return os;
} // operator<<


//
//	------------------( RightAscension / Declination )------------------------
//
namespace {
std::vector<double> breakit(const std::string &toBreakString, bool ra = true)
{
	std::vector<std::string> v(breakup(toBreakString, ": dhmso'\""));

	if (v.size() == 0 || v.size() > 3)
		errstr<<"Cannot convert Right-Acension/Declination specification "
				"string `"<<toBreakString<<"' to radians"
			  <<Sas::error("WrongFormat");

	std::vector<double> d;
	d.push_back(v[0][0] == '-' ? -1. : 1.);	//	sign

	try {

	if (v.size() == 1)
		if (ra)
			d.push_back(std::fabs(toNumber<double>(v[0])));	//	result in hours
		else
			d.push_back(std::fabs(toNumber<double>(v[0])));	//	result in deg

	for (unsigned n=v.size(); n<3; ++n)
		v.push_back("0");

	for (unsigned n=d.size(); n<v.size()+1; ++n)
		d.push_back(toNumber<double>(v[n-1]));

	} catch (string & e) {
	  errstr << "In Celestial.cc I could not convert a string to a double: "
			 << e << Sas::error("FailedConvertion");
	} catch(...) {
	  errstr << "In Celestial.cc I caught an unknown exception, probably while trying to convert a string to double." << Sas::error("FailedConvertion");
	}

	return d;
} // breakit()
}


RightAscension::RightAscension(const std::string &raString)
{
	std::vector<double> dv(breakit(raString));

	_ra = dv[0]*(15. * std::fabs(dv[1]) + .25 * dv[2] + dv[3] / 240.) * DEG_TO_RAD;

} // RightAscension::RightAscension()


std::string RightAscension::toString(char sep) const
{
	//
	//	normalize alpha: 0<= alpha <= 360
	//
	double alpha = std::fmod(_ra * RAD_TO_DEG, 360.);

    if (alpha < 0.)
        alpha += 360.;

	alpha = DBL(UNT(240000. * alpha + .5) / 240000.);

	double hours, mins;
	double secs = 3600. * std::modf(240./3600. * alpha, &hours);
	secs = 60. * std::modf(secs/60., &mins);
	unsigned fracSecs = UNT(100. * std::modf(secs, &secs)+.5);
	if (fracSecs == 100)
		fracSecs = 99;

	stringstream s; s.precision(0); s.fill('0');
	if (!sep)
		s<<setw(2)<<INT(hours)<<'h'<<setw(2)<<INT(mins)<<'m'<<setw(2)
		 <<INT(secs)<<'.'<<setw(2)<<setprecision(0)<<fracSecs<<'s';
	else
		s<<setw(2)<<INT(hours)<<sep<<setw(2)<<INT(mins)<<sep<<setw(2)
		  <<INT(secs)<<'.'<<setw(2)<<setprecision(0)<<fracSecs;

	return s.str();
} // RightAscension::toString()


Declination::Declination(const string &decString)
{
	vector<double> dv(breakit(decString, false));

	_dec = dv[0] * (std::fabs(dv[1]) + dv[2]/60. + dv[3]/3600.) * DEG_TO_RAD;
} // Declination::Declination()


std::string Declination::toString(char sep) const
{
	double delta = (std::fmod(_dec+M_HALF_PI, M_PI)-M_HALF_PI) * RAD_TO_DEG;
	bool negDec = delta > 0. ? false : true;
	delta = DBL(UNT(360000. * std::fabs(delta) + .5) / 360000.);

	double degs, mins;
	double secs = 3600. * std::modf(delta, &degs);
	secs = 60. * std::modf(secs/60., &mins);
	unsigned fracSecs = UNT(100. * std::modf(secs, &secs)+.5);
	if (fracSecs == 100)
		fracSecs = 99;

	stringstream s; s.precision(0); s.fill('0');

	if (!negDec) {
		if (!sep)
			s<<setw(2)<<INT(degs)<<'o'<<setw(2)<<INT(mins)<<'\''
		 	<<setw(2)<<INT(secs)<<'.'<<setw(2)<<INT(fracSecs)<<"''";
		else
			s<<setw(2)<<INT(degs)<<sep<<setw(2)<<INT(mins)<<sep
			 <<setw(2)<<INT(secs)<<'.'<<setw(2)<<INT(fracSecs);
	}
	else {
                if (!sep)
                        s<<'-'<<setw(2)<<INT(degs)<<'o'<<setw(2)<<INT(mins)<<'\''
                        <<setw(2)<<INT(secs)<<'.'<<setw(2)<<INT(fracSecs)<<"''";
                else
                        s<<'-'<<setw(2)<<INT(degs)<<sep<<setw(2)<<INT(mins)<<sep
                         <<setw(2)<<INT(secs)<<'.'<<setw(2)<<INT(fracSecs);
	}

	return s.str();
} // Declination::toString()


EquatorialDirection::EquatorialDirection(const CartesianVector &v,
										 double epoch)
	: _epoch(epoch)
{
	CartesianVector r(v);
	r.normalize();

	_dec = Declination(std::asin(r.z()));
	_ra = RightAscension(r.x() != 0. ||
						 r.y() != 0. ? std::atan2(r.y(), r.x()) :0.);
	if (_ra.angle() < 0.)
		//	result of atan2 is in intervals [-Pi, Pi]
		_ra += M_TWO_PI;
} // EquatorialDirection::EquatorialDirection()


CartesianVector EquatorialDirection::toUnitCartesian() const
{
	double cosDec = std::cos(_dec.angle());
	CartesianVector vec(std::cos(_ra.angle()) * cosDec,
					    std::sin(_ra.angle()) * cosDec,
						std::sin(_dec.angle()));

	return vec;
} // EquatorialDirection::toUnitCartesian


EquatorialDirection &EquatorialDirection::toEpoch(double epoch)
{
	if (epoch != _epoch)
		errstr<<"Precession of celestial coordinates from epoch "
			  <<_epoch<<" to "<<epoch<<" not implemented yet."
			  <<Sas::warning("NotImplementedYet");

	return *this;
} // EquatorialDirection::toEpoch()


EquatorialDirection	&EquatorialDirection::toApparentDirection(
													const CartesianVector &r,
													double sourceDist)
{
	EquatorialDirection apparentDir(sourceDist * toUnitCartesian() - r,
									_epoch);
	_ra = apparentDir.ra();
	_dec = apparentDir.dec();

	return *this;
} // EquatorialDirection::toApparentDirection()


//
//	---( WCS stuff )-----------------------------------------------------------
//
namespace WCS {

TangentialGnomonicProjection::TangentialGnomonicProjection(
								const EquatorialDirection &refPoint,
								double refX, double deltaX,
								double refY, double deltaY,
								double longPole)
	: _raRef(refPoint.ra().angle()), _decRef(refPoint.dec().angle()),
	  _refX(refX), _deltaX(deltaX), _refY(refY), _deltaY(deltaY),
	  _longPole(longPole)
{
	_sinDecRef = std::sin(_decRef);
	_cosDecRef = std::cos(_decRef);
} // TangentialGnomonicProjection::TangentialGnomonicProjection()


EquatorialDirection TangentialGnomonicProjection::toStandardSystem(
											const NativePlaneCoordinates &pos)
{
	double x = _deltaX * (pos.first - _refX),
		   y = _deltaY * (pos.second - _refY);

	double Rtheta = sqrt(x*x + y*y),
		   theta = M_HALF_PI,
		   phi = 0.;
	if (Rtheta != 0.) {
		theta = std::atan(1./Rtheta);
		phi = std::atan2(x, -y);
	}
	SAS_DEBUG(cout<<"theta/phi="<<theta*RAD_TO_DEG<<' '<<phi*RAD_TO_DEG<<std::endl;)

	double delta = std::asin(std::sin(theta)*_sinDecRef -
					  std::cos(theta)*_cosDecRef*std::cos(phi)),
		   alpha = _raRef + std::asin(std::cos(theta)*std::sin(phi)/
						std::cos(delta));

	return EquatorialDirection(RightAscension(alpha), Declination(delta));
} // TangentialGnomonicProjection::toStandardSystem()


ZenithalProjection::NativePlaneCoordinates
TangentialGnomonicProjection::toNativeSystem(const EquatorialDirection &dir)
{
	double alpha = dir.ra().angle(),
		   delta = dir.dec().angle(),
		   theta = std::asin(std::sin(delta)*_sinDecRef +
						std::cos(delta)*_cosDecRef*std::cos(alpha-_raRef)),
		   arg0 = -std::cos(delta)*std::sin(alpha-_raRef),
		   arg1 = std::sin(delta)*_cosDecRef -
					 std::cos(delta)*_sinDecRef*std::cos(alpha-_raRef),
		   phi   = std::fabs(arg0)>DBL_MIN || std::fabs(arg1)>DBL_MIN
					? _longPole + std::atan2(arg0, arg1) : 0.,

		   Rtheta = 1./std::tan(theta),
		   x = Rtheta * std::sin(phi),
		   y = -Rtheta * std::cos(phi);

	return NativePlaneCoordinates(x/_deltaX + _refX, y/_deltaY + _refY);
} // TangentialGnomonicProjection::toNativeSystem()

} // namespace WCS
