//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	EarthPositionLocator.h --- inquire position of object on the earth
//
//	DESCRIPTION:
//		An EarthPositionLocator object can track the position of a fixed
//		object on the surface of the earth.
//
//	CHANGE HISTORY:
//		UL, 2000-11-02: 1.0.0: - first version
//		UL, 2003-02-20: 1.1.0: - changed name of private variable to avoid
//								 name clash under MacOSX
//
//	AUTHOR:
//		U. Lammers, Thu Nov  2 16:57:54 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: "Fundamental Astronomy", H. Karttunen et al., Springer-Verlag
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/caloalutils/src/EarthPositionLocator.h,v 1.8 2003/02/20 14:36:40 ulammers Exp $
//

#ifndef EARTH_POSITION_LOCATOR_H
#define EARTH_POSITION_LOCATOR_H

#include "Celestial.h"
#include "STime.h"
#include "SasException.h"
#include "PositionLocator.h"

//
//	A fixed object on the surface of the earth
//
class EarthObject {
public:
	//
	//	exceptions
	//
	class BadLattitude : public SasException::Exception {
	public:
		BadLattitude()
			: SasException::Exception("Given Lattitude outside allowed range "
									  "[-Pi/2, Pi/2]") {}
	};

	enum LattitudeType { GEODETIC, GEOCENTRIC };
	EarthObject(double longitude = 0.,
				double lattitude = 0., LattitudeType lattType = GEODETIC,
				double altitude = 0,
				const std::string &name = "Unknown") throw(BadLattitude);
		//
		//	PURPOSE:
		//		Construct an object from its longitude/lattitude on the
		//		surface of the Earth
		//
		//	ARGUMENTS:
		//		longitude:	geographic longitude, 0=Greenwich Meridian,
		//					positively increasing eastwards [rad]
		//		lattitude:  geodetic/geographic lattitude, 0=equator,
		//					>0 in Northern-/<0 in Southern hemisphere [rad]
		//		altitude :  altitude of the object above NN [m]
		//


	double longitude() const throw() { return _long; }
	double lattitude(LattitudeType lattType = GEODETIC) const throw()
	{ return lattType == GEODETIC ? _lattGeoDet : _lattGeoCen; }
	double altitude() const throw() { return 1.e3*_altitude; }
	const std::string &name() const { return _name; }
		//
		//	PURPOSE:
		//		Returns longitude [rad]/lattidue [rad]/altitude [m]/name of
		//		object;
		//


	const CartesianVector &position() const throw() { return _pos; }
		//
		//	PUPROSE:
		//		Return position vector of object in Cartesian earth-centered
		//		reference frame, with
		//			+Z axis pointing to North
		//			+X axis pointing to equator in the direction of
		//			 Zero-Meridian
		//			+Y completes right handed triad
		//


	~EarthObject() {}

private:
	//	data
	double		_long,
				_lattGeoCen,	//	geocentric lattitude
				_lattGeoDet,	//	geodetic lattitude
				_altitude;		//	altitude above NN [km]
	std::string	_name;			//	name of the place

	CartesianVector	_pos;	//	position vector

	static const double polarR = 6356.755;	// [km], see [2]
	static const double equatR = 6378.140;	// [km], see [2]

	static const double polarRsquare = polarR * polarR;
	static const double equatRsquare = equatR * equatR;
};
	

class EarthPositionLocator : public PositionLocator {
public:
	EarthPositionLocator(const EarthObject &observer,
						 const STime &time = STime("2000-01-01")) throw()
		: PositionLocator(time), _observer(observer),
		  _rotdist(sqrt(observer.position().x()*observer.position().x()+
					    observer.position().y()*observer.position().y())) {}
		//
		//	PURPOSE:
		//		Construct EarthPositionLocator from EarthObject and
		//		optional base time
		//


	const EarthObject &earthObject() const { return _observer; }
		//
		//	PURPOSE:
		//		Return EarthObject of this PositionLocator
		//


	StateVector stateVector(double time) const throw();
	StateVector stateVector(const STime &time) const throw();
		//
		//	PURPOSE:
		//		Return state vector of object in Earth-centered, Cartesian
		//		J2000 reference frame at given time
		//

	~EarthPositionLocator() {}

private:
	EarthObject	_observer;
	double		_rotdist;	//	distance from Earth rotational axis [km]
};

#endif
