//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997 - 2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	KeplerOrbitSimulator.cc --- implementation of KeplerOrbitSimulator
//
//	DESCRIPTION:
//		This is a simple hack to mimic the behavior of a rigid small body which
//		moves in a Keplerian orbit around a massive second body.
//
//	CHANGE HISTORY:
//		UL, 1999-06-11: 1.0.0: - first version
//		UL, 1999-06-15: 1.1.0: - minor bug fix
//							   - added rudimentary calculation of velocity
//								 vetor
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/caloalutils/src/KeplerOrbitSimulator.cc,v 1.5 2003/05/23 14:33:37 ulammers Exp $
//

#include <cmath>
#include "KeplerOrbitSimulator.h"

#include "errstr.h"


KeplerOrbitSimulator::KeplerOrbitSimulator(
							const OrbitalElements &orbitalElements,
							const double initialPerigeePassTime)
		: _orbitalElements(orbitalElements),
		  _initialPerigeePassTime(initialPerigeePassTime),
		  _orbitalElementsSet(true)
{
	init();
} // KeplerOrbitSimulator::KeplerOrbitSimulator()


void KeplerOrbitSimulator::init()
{
	AttitudeMatrix a(_orbitalElements.ascendingNode,
					 _orbitalElements.inclincation,
					 -(2.*M_HALF_PI-_orbitalElements.perigeeArgument));

	_a313 = a.inverse();	//	we need the inverse to go from rotated
							//	frame into original frame
	_omega = M_TWO_PI/_orbitalElements.period;
	_a = EARTH_RADIUS +
			.5 * (_orbitalElements.apogeeHeight +
					_orbitalElements.perigeeHeight);
	_c = .5 * (_orbitalElements.apogeeHeight - _orbitalElements.perigeeHeight);
	_b = _a * sqrt(1.-_orbitalElements.eccentricity*
					  _orbitalElements.eccentricity);

//	_phi0 = M_PI - _omega * _initialPerigeePassTime;
} // KeplerOrbitSimulator::init()


CartesianVector KeplerOrbitSimulator::position(double time) const
{
	if (!_orbitalElementsSet)
		errstr<<"Orbital elements for this simulator have not been set - "
				"cannot continue"<<Sas::error("ElementsUndefined");

	double phi = std::fmod(_omega *
							(time - _initialPerigeePassTime), M_TWO_PI),
		   x = _a * std::cos(phi),
		   y = _b * std::sqrt(1. - x*x/(_a*_a));

	if (phi > M_PI)
		y = -y;

	CartesianVector v(_c + x, y, 0.);

	return _a313 * v;
} // KeplerOrbitSimulator::position()


CartesianVector KeplerOrbitSimulator::velocity(double time) const
{
	return .1 * (position(time + 10.) - position(time));
} // KeplerOrbitSimulator::velocity()


void KeplerOrbitSimulator::orbitalElements(
								const OrbitalElements &orbitalElements,
								double initalPerigeePassTime)
{
	_orbitalElements = orbitalElements;
	_orbitalElementsSet = true;
	_initialPerigeePassTime = initalPerigeePassTime;

	init();
} // KeplerOrbitSimulator::orbitalElements()
