//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	RealCcf.cc	--- interface between the CAL and the CIF in the CCF
//
//	DESCRIPTION:
//		This defines the interface between the CAL and the CIF in the CCF
//
//	CHANGE HISTORY:
//		UL, 1997-10-20: 1.0.0: - skeleton
//		UL, 1999-02-09: 1.1.0: - name of environment variable pointing to
//								 CCF directory changed
//		UL, 1999-03-05: 1.2.0: - changes pertaining to lowmem Dal
//		UL, 1999-03-17: 1.3.0: - d/f name of CIF changed to "ccf.cif"
//		UL, 1999-05-12: 1.3.1: - changed fatal -> error and changed all
//								 error/warning tags
//		UL, 1999-10-11: 1.3.2: - only prepend SAS_CCFDIR to constituent file
//								 name if not absolute (fix SPR 889:
//								 Local replacement of a CCF does not work
//		UL, 1999-11-30: 1.4.0: - dropped support for old SAS_CCFDIR
//		UL, 2000-04-23: 1.5.0: - open CIF always in low-mem mode
//		UL, 2003-05-23: 1.6.0: - added RealCcf::ccfReplacements(): inspects
//								 enviroment variable SAS_CCFFILES for
//								 any given CCF replacements (see CAL doc
//								 for more details)
//
//	AUTHOR:
//		U. Lammers, Tue Oct 14 14:45:35 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: Indexing of the XMM Current Calibration File, U. Lammers,
//			 15/07/97
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/caloalutils/src/RealCcf.cc,v 1.3 2003/05/23 14:33:37 ulammers Exp $
//
#include <cstdlib>

#include "RealCcf.h"
#include "Dal.h"
#include "CalOalUtility.h"
#include "stringutil.h"

using namespace std;

const string CalIndexExt= "CALINDEX";	// extension name for the index table

//
//	the constructor: receives the name of the CIF and initializes
//	private member functions by opening the CIF and populating the
//	CIF map accordingly.
//
RealCcf::RealCcf(const string cifFileName) : Ccf(cifFileName)
{
	using namespace CalOalUtility;

	// open the CIF
	errstr<<"Opening CCF CIF "<<cifName()<<endl
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	DataSet	*cif = dataSetServer->open(cifName(), DataSet::Read,
									   DataSetServer::Low);
	Table	*cifTab = cif->table(CalIndexExt);

	// get column numbers
	string colNames[] = { "SCOPE", "TYPEID", "ISSUE", "FNAME", "VALDATE",
						  "DATE", "EXTSEQU", "EXTSEQID" };

	ColumnData *cd[NUM_OF(colNames)];
	for (unsigned i=0; i<NUM_OF(colNames); ++i)
		cd[i] = cifTab->column(colNames[i])->data();

	//	loop over all rows
	for (unsigned n=0; n<cifTab->rows(); ++n) {
		string instVal = stringColumnRow(cd[0], n),
			   typeidVal = stringColumnRow(cd[1], n),
			   fnameVal = stringColumnRow(cd[3], n),
			   valdateVal = stringColumnRow(cd[4], n),
			   subdateVal = stringColumnRow(cd[5], n),
			   extSeqVal = stringColumnRow(cd[6], n),
			   extSeqIdVal = stringColumnRow(cd[7], n);
		int16  issueVal = cd[2]->int16Data()[n];

		Xmm::Instrument instr = Xmm::toInstrument(instVal);
		Ccf::CalDataTypes type  = Ccf::toTypeId(typeidVal);

		if (instr < 0 || type < 0) {
			// bail-out - illegal instrument/data type name
			errstr<<instVal<<"/"<<typeidVal<<" is not a known "
					"instrument/filetype combination"
				  <<Sas::warning("Mismatch");
		    continue;
		}

		CcfConstituent constit(instr, type);
		CifEntry *e	= new CifEntry(instr,
								   type,
								   issueVal,
								   fnameVal,
								   valdateVal,
								   subdateVal,
								   extSeqVal,
								   extSeqIdVal, this);

		errstr<<"New Cif entry: "<<*e<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		//	insert CIF entry into map
		_cif[constit] = e;
	}

	//	tidy up release data set
	for (unsigned n=0; n<NUM_OF(colNames); ++n)
		delete cd[n];
	dataSetServer->close(cif);

	//	inspect enviorment for CCF replacements
	ccfReplacements();
} // RealCcf::RealCcf()


void RealCcf::ccfReplacements()
{
	char *ccfFiles = getenv(Ccf::SAS_CCFFILES);
	if (ccfFiles) {
		vector<string> items(breakup(ccfFiles));
		for (unsigned n=0; n<items.size(); ++n)
			registerNewCifEntry(items[n]);
	}
} // RealCcf::ccfReplacements()


//	destructor - be nice and free-up some memory
RealCcf::~RealCcf()
{
	for (Cif::const_iterator i = _cif.begin(); i != _cif.end(); ++i)
		delete (*i).second;
} // RealCcf::~RealCcf()


//	register new CIF entry
void RealCcf::registerNewCifEntry(const std::string &setName)
{
	CifEntry entry(setName, this);
	CcfConstituent constit(entry.instrument(), entry.calDataType());
	Cif::const_iterator i = _cif.find(constit);

	if (i != _cif.end()) {
		errstr<<"Registering new calibration file: "
			  <<Xmm::toString(entry.instrument())
			  <<"/"<<Ccf::toString(entry.calDataType())
			  <<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);
		delete (*i).second;
	}

	_cif[constit] = new CifEntry(entry);
} // RealCcf::registerNewCifEntry()


//	query for a CIF entry
CifEntry *RealCcf::entry(const XmmInstrument *instr, Ccf::CalDataTypes type,
						 bool generateError)
const
{
	CcfConstituent	constit(instr->id(), type);
	Cif::const_iterator i = _cif.find(constit);

	errstr<<"Querying CIF for CCF constituent: "
		  <<Xmm::toString(instr->id())<<"/"<<Ccf::toString(type)<<endl
		  <<Sas::message(Msg::Lib, Msg::Verbose);
	if (i == _cif.end())
		//	not found - try the corresponding telescope file
		if ((i = _cif.find(CcfConstituent(instr->behindTelescope(), type)))
			== _cif.end())
			//	hmm, let's try the satellite version then
			if ((i=_cif.find(CcfConstituent(Xmm::SC, type))) == _cif.end())
				//	something is definitely wrong
				if (generateError)
					errstr<<"The CCF does not contain a "
						  <<Xmm::toString(instr->id())<<"/"
						  <<Ccf::toString(type)<<" file - also tried "
						  "the version in the telescope and satellite scope"
						  <<Sas::error("DataSetNotFound");
				else {
					return 0;
				}

	return (*i).second;
} // RealCcf::CifEntry


bool RealCcf::hasEntry(const XmmInstrument *instr, Ccf::CalDataTypes type)
const
{
	return entry(instr, type, false);
} // RealCcf::hasEntry()



void RealCcf::print(ostream &os) const
{
	os<<"Ccf{"<<_cifFileName<<endl;
	for (Cif::const_iterator i = _cif.begin();
		 i!=_cif.end(); ++i) {
		os<<"   [("<<Xmm::toString((*i).first.first)<<","
		  <<Ccf::toString((*i).first.second)
		  <<"), "<<*(*i).second<<"]\n";
	}
	os<<"}";
} // RealCcf::print()



